from collections import defaultdict

from django.utils.module_loading import import_string

from rest_framework.metadata import SimpleMetadata, BaseMetadata

from .utils import get_languages, get_field_dict
from .app_settings import settings
from .adapters import GETTER


class AutoMetadataMixin(object):

    def root_metadata(self, metadata, view):
        from .router import router
        rv = {
            'endpoints': [k for k in router._endpoints.keys()]
        }

        applications = defaultdict(lambda: [])
        for url, endpoint in router._endpoints.items():
            if endpoint.list_me:
                applications[endpoint.application_name].append({
                    'name': endpoint.model_name,
                    'singular': endpoint.singular_model_name,
                    'endpoint': url
                })
        rv['applications'] = [
            {
                'name': k,
                'models': v
            } for k, v in applications.items()

        ]
        rv['languages'] = get_languages()

        adapter = import_string(settings.METADATA_ADAPTER)()
        metadata.update(adapter.render_root(rv))
        return metadata

    def determine_metadata(self, request, view):

        try:
             metadata = super(AutoMetadataMixin, self).determine_metadata(request, view)
        except NotImplementedError:
            metadata = {}
        except AttributeError:
            metadata = {}

        root_view_names = ['APIRootView', 'APIRoot']
        if view.__class__.__name__ in root_view_names or view in root_view_names:
            return self.root_metadata(metadata, view)

        serializer_instance = view.serializer_class()
        endpoint = None
        if hasattr(view, 'endpoint'):
            endpoint = view.endpoint
        else:
            if hasattr(view.serializer_class.Meta, 'model'):
                from .endpoints import Endpoint
                endpoint = Endpoint(view.serializer_class.Meta.model, viewset=view)

        adapter = import_string(settings.METADATA_ADAPTER)()
        if endpoint is None:
            fields_metadata = []

            for field in view.serializer_class.Meta.fields:
                if field in {'id', '__str__'}:
                    continue

                instance_field = serializer_instance.fields[field]
                type_ = settings.WIDGET_MAPPING.get(instance_field.__class__.__name__)

                if type_ is None:
                    raise NotImplementedError()

                field_metadata = get_field_dict(field, view.serializer_class)

                fields_metadata.append(field_metadata)

                for meta_info in adapter.metadata_info:
                    if meta_info.attr == 'fields':
                        metadata['fields'] = fields_metadata,
                    elif meta_info.attr == 'fieldsets':
                        metadata['fieldsets'] = [{
                            'title': None,
                            'fields': [
                                {'key': field}
                                for field in view.serializer_class.Meta.fields
                                if field != 'id' and field != '__str__'
                            ]
                        }]
                    else:
                        metadata[meta_info.attr] = meta_info.default
        else:
            for meta_info in adapter.metadata_info:
                try:
                    if meta_info.attr_type == GETTER:
                        method = getattr(endpoint, 'get_{}'.format(meta_info.attr))
                        try:
                            metadata[meta_info.attr] = method(request)
                        except TypeError:
                            metadata[meta_info.attr] = method()
                    else:
                        metadata[meta_info.attr] = getattr(endpoint, meta_info.attr, meta_info.default)
                except AttributeError:
                    metadata[meta_info.attr] = meta_info.default

        return adapter(metadata)


class AutoMetadata(AutoMetadataMixin, SimpleMetadata):
    pass


class MinimalAutoMetadata(AutoMetadataMixin, BaseMetadata):
    pass


class RootViewMetadata(SimpleMetadata):
    pass



from collections import OrderedDict

from django.utils.module_loading import import_string

from rest_framework.routers import DefaultRouter

from .endpoints import Endpoint
from .app_settings import settings


class EndpointRouter(DefaultRouter):

    base_endpoint_class = Endpoint

    def __init__(self, *args, **kwargs):
        self._endpoints = OrderedDict()
        self._registry = {}
        super(EndpointRouter, self).__init__(*args, **kwargs)

    def register(self, model=None, endpoint=None, fields=None, permission_classes=None,
                 serializer=None, filter_fields=None, read_only=False, viewset=None,
                 search_fields=None, ordering_fields=None, page_size=None, base_viewset=None,
                 base_name=None, fields_annotation=None, fielsets=None, base_serializer=None,
                 list_me=True, **kwargs):

        if (endpoint is None or isinstance(endpoint, type)):
            extra = {}
            if base_viewset is not None:
                extra['base_viewset'] = base_viewset
            if base_serializer is not None:
                extra['base_serializer'] = base_serializer

            endpoint_kwargs = {
                'model': model,
                'fields': fields,
                'fieldsets': fielsets,
                'permission_classes': permission_classes,
                'serializer': serializer,
                'filter_fields': filter_fields,
                'read_only': read_only,
                'viewset': viewset,
                'search_fields': search_fields,
                'ordering_fields': ordering_fields,
                'fields_annotation': fields_annotation,
                'list_me': list_me
            }
            endpoint_kwargs.update(extra)

        if endpoint is None:
            endpoint = self.base_endpoint_class(**endpoint_kwargs)
        elif isinstance(endpoint, type):
            endpoint = endpoint(**endpoint_kwargs)

        url = endpoint.get_url() if 'url' not in kwargs else kwargs.pop('url')
        self._endpoints[url] = endpoint

        if base_name is None:
            base_name = url

        super(EndpointRouter, self).register(
            url,
            endpoint.get_viewset(),
            base_name=base_name,
            **kwargs
        )

    def override_registry_entry(self, endpoint):
        url = endpoint.get_url()
        self._endpoints[url] = endpoint

        new_registry = []
        for (prefix, viewset, base_name) in self.registry:
            if prefix == url:
                new_registry.append((url, endpoint.get_viewset(), base_name))
            else:
                new_registry.append((prefix, viewset, base_name))
        self.registry = new_registry

    def get_endpoint(self, url):
        return self._endpoints[url]

    def registerViewSet(self, *args, **kwargs):
        super(EndpointRouter, self).register(*args, **kwargs)


def register(wrapped=None, **kwargs):
    from drf_auto_endpoint.router import router as default_router

    def _endpoint_wrapper(endpoint_class):
        router = kwargs.pop('router', default_router)
        router.register(endpoint=endpoint_class(), **kwargs)
        return endpoint_class

    if wrapped is not None:
        return _endpoint_wrapper(wrapped)
    return _endpoint_wrapper


router = import_string(settings.ROUTER_CLASS)()
